@tool
extends EditorPlugin

var button : Control
var eds = get_editor_interface().get_selection()

var selected_node

var fd : FileDialog

var finger_names : Array = [
	"Thumb",
	"Index",
	"Middle",
	"Ring",
	"Pinky"
]

func _enter_tree():
	# Initialization of the plugin goes here.
	# Load the dock scene and instantiate it.
	button = preload("res://addons/UltraleapHandMenu/Button.tscn").instantiate()
	button.visible = false
	button.get_popup().id_pressed.connect(_on_id_pressed)
	
	fd = button.get_node("FileDialog")
	fd.file_mode = FileDialog.FILE_MODE_OPEN_FILES
	fd.access = FileDialog.ACCESS_RESOURCES
	fd.files_selected.connect(_on_file_selected)

	# Add the loaded scene to the docks.
	add_control_to_container(EditorPlugin.CONTAINER_SPATIAL_EDITOR_MENU, button)
	# Note that LEFT_UL means the left of the editor, upper-left dock.
	
	eds.selection_changed.connect(_on_selection_changed)

func _exit_tree():
	# Clean-up of the plugin goes here.
	# Remove the dock.
	remove_control_from_container(EditorPlugin.CONTAINER_SPATIAL_EDITOR_MENU, button)
	# Erase the control from the memory.
	button.free()


func _on_selection_changed():
	# Returns an array of selected nodes
	var selected = eds.get_selected_nodes() 

	if selected.size() == 1:
		# Always pick first node in selection
		if selected[0] is DebugHands:
			selected_node = selected[0]
			button.visible = true
		else:
			button.visible = false

func _on_id_pressed(id : int):
	if id == 0:
		print("Create hands")
		create_hand(UltraleapTypes.Chirality.Left, selected_node)
		create_hand(UltraleapTypes.Chirality.Right, selected_node)
	if id == 1:
		print("Apply pose")
		fd.popup_centered()

func create_hand(chirality : UltraleapTypes.Chirality, parent : Node):
	var prefix : String = "Left" if chirality == UltraleapTypes.Chirality.Left else "Right"

	var hand = Node3D.new()
	hand.name = prefix + "Hand"
	
	parent.add_child(hand)
	hand.set_owner(get_tree().get_edited_scene_root())
	
	var palm = Node3D.new()
	palm.name = prefix + "Palm"
	
	#var arm = MeshInstance3D.new()
	#arm.mesh = BoxMesh.new()
	#arm.name = prefix + "Arm"
	
	var arm = create_bone(prefix, "Arm")
	hand.add_child(arm)
	arm.set_owner(get_tree().get_edited_scene_root())
	
	hand.add_child(palm)
	palm.set_owner(get_tree().get_edited_scene_root())

	create_finger(prefix, "Thumb", palm)
	create_finger(prefix, "Index", palm)
	create_finger(prefix, "Middle", palm)
	create_finger(prefix, "Ring", palm)
	create_finger(prefix, "Pinky", palm)
	
	return hand

func create_finger(prefix : String, type : String, parent : Node):
	prefix = prefix + type

	var metacarpal = create_bone(prefix, "Metacarpal")
	var proximal = create_bone(prefix, "Proximal")
	var intermediate = create_bone(prefix, "Intermediate")
	var distal = create_bone(prefix, "Distal")
	
	parent.add_child(metacarpal)
	metacarpal.set_owner(get_tree().get_edited_scene_root())
	
	metacarpal.add_child(proximal)
	proximal.set_owner(get_tree().get_edited_scene_root())
	
	proximal.add_child(intermediate)
	intermediate.set_owner(get_tree().get_edited_scene_root())
	
	intermediate.add_child(distal)
	distal.set_owner(get_tree().get_edited_scene_root())
	
	return metacarpal

func create_bone(prefix : String, name : String):
	var bone = MeshInstance3D.new()
	bone.mesh = CapsuleMesh.new()
	bone.name = prefix + name
	bone.mesh.material = StandardMaterial3D.new()
	bone.mesh.radial_segments = 8
	bone.mesh.rings = 4
	return bone

func _on_file_selected(paths):
	print(paths)
	for path in paths:
		var file = load(path)
		if file is UltraleapHand:
			var hand : UltraleapHand = file as UltraleapHand
			set_hand_transform(hand, hand.type)
		
func set_hand_transform(hand : UltraleapHand, chirality : UltraleapTypes.Chirality):
	var chirality_str = chirality_to_str(chirality)
	var hand_root = selected_node.get_node(chirality_str + "Hand") as Node3D
	
	var arm = hand_root.get_node(chirality_str + "Arm")
	
	set_bone_transform(hand.arm, arm)
	
	for i in range(5):
		set_finger_transform(hand.digits[i], chirality, finger_names[i], hand_root)

func set_finger_transform(digit : UltraleapDigit, chirality : UltraleapTypes.Chirality, finger : String, node : Node3D):
	var chirality_str = chirality_to_str(chirality)
	var palm = node.get_node(chirality_str + "Palm")
	var metacarpal = palm.get_node(chirality_str + finger + "Metacarpal")
	var proximal = metacarpal.get_node(chirality_str + finger + "Proximal")
	var intermediate = proximal.get_node(chirality_str + finger + "Intermediate")
	var distal = intermediate.get_node(chirality_str + finger + "Distal")
	
	set_bone_transform(digit.metacarpal, metacarpal)
	set_bone_transform(digit.proximal, proximal)
	set_bone_transform(digit.intermediate, intermediate)
	set_bone_transform(digit.distal, distal)
	
func set_bone_transform(bone : UltraleapBone, model : Node3D):
	var pos = (bone.next_joint + bone.prev_joint) / 2
	var length = (bone.next_joint - bone.prev_joint).length()
	var transform_m : Transform3D = Transform3D(selected_node.global_transform)

	model.global_position = transform_m * pos
	
	# Multiply by quaternion, which is this object rotation, to rotate the hands
	model.global_rotation = (selected_node.quaternion * bone.rotation * Quaternion(Vector3.RIGHT, deg_to_rad(90))).get_euler()
	
	var mesh
	var material

	if model is CSGBox3D:
		mesh = model
		material = (mesh.material as StandardMaterial3D)
	if model is  MeshInstance3D:
		mesh = model.mesh
		material = model.mesh.material
		
	if mesh is CapsuleMesh:
		mesh.radius = bone.width / 2
		mesh.height = length
	else:
		mesh.size.x = bone.width
		mesh.size.z = length
		mesh.size.y = 0.01

func chirality_to_str(chirality : UltraleapTypes.Chirality):
	return "Left" if chirality == UltraleapTypes.Left else "Right"
