#include "device_node.h"

#include <godot_cpp/core/class_db.hpp>

#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/classes/ref.hpp>
#include <godot_cpp/classes/node3d.hpp>

#include <godot_cpp/variant/utility_functions.hpp>

#include <LeapC.h>

#include "ultraleap.h"

using namespace godot;

void UltraleapDeviceNode::device_added(Ref<UltraleapDevice> device) {
    if (!devices.has(device->serial)) {
        devices.append(device->serial);
        notify_property_list_changed();
    }

    // If user has chosen "Any", the strategy here is to use the newly connected device
    if (chosen_serial == 0) {
        if (current_device != NULL) {
            unsubscribe_and_close();
        }
        current_device = device;
        open_and_subscribe();
    }
    // Otherwise we check for the serial and subscribe to the device
    else {
        if (get_tracking() != NULL) {
            if ((chosen_serial - 1) < get_tracking()->devices->size()) {
                if (devices[chosen_serial - 1] == device->serial) {
                    current_device = device;
                    open_and_subscribe();
                }
            }
        }
    }
}

void UltraleapDeviceNode::device_removed(Ref<UltraleapDevice> device) {
    if (!devices.has(device->serial)) {
        return;
    }

    // If the device being removed is the same as the one being used by this
    // node, unsubscribe and close it
    if (current_device != NULL) {
        if (device->serial == current_device->serial) {
            unsubscribe_and_close();
            current_device = Ref<UltraleapDevice>();
        }
    }

    devices.erase(device->serial);

    notify_property_list_changed();
}

void UltraleapDeviceNode::open_and_subscribe() {
    if (current_device != NULL) {
        current_device->open();
        current_device->subscribe();

        current_device->connect("tracking_mode_changed", Callable((Object*)this, "on_tracking_mode_changed"));
        current_device->connect("frame_received", Callable((Object*)this, "on_tracking_event_received"));
    }
}

void UltraleapDeviceNode::unsubscribe_and_close() {
    if (current_device != NULL) {
        current_device->unsubscribe();
        current_device->close();

        current_device->disconnect("tracking_mode_changed", Callable((Object*)this, "on_tracking_mode_changed"));
        current_device->disconnect("frame_received", Callable((Object*)this, "on_tracking_event_received"));
    }
}

void UltraleapDeviceNode::_bind_methods() {
    PropertyInfo tracking_mode_info = PropertyInfo(
        Variant::INT,
        "tracking_mode",
        PROPERTY_HINT_ENUM,
        "Desktop, HMD, Screentop"
    );

    ClassDB::bind_method(D_METHOD("set_tracker", "tracker"), &UltraleapDeviceNode::set_tracker);
    ClassDB::bind_method(D_METHOD("get_tracker"), &UltraleapDeviceNode::get_tracker);

    ClassDB::bind_method(D_METHOD("get_last_frame"), &UltraleapDeviceNode::get_last_frame);

    ClassDB::bind_method(D_METHOD("device_added", "device"), &UltraleapDeviceNode::device_added);
    ClassDB::bind_method(D_METHOD("device_removed", "device"), &UltraleapDeviceNode::device_removed);

    ClassDB::bind_method(D_METHOD("set_tracking_mode", "tracking_mode"), &UltraleapDeviceNode::set_tracking_mode, DEFVAL(0));
    ClassDB::bind_method(D_METHOD("get_tracking_mode"), &UltraleapDeviceNode::get_tracking_mode);

    ClassDB::bind_method(D_METHOD("request_tracking_mode", "tracking_mode"), &UltraleapDeviceNode::request_tracking_mode, DEFVAL(0));
    ClassDB::bind_method(D_METHOD("on_tracking_mode_changed", "tracking_mode"), &UltraleapDeviceNode::on_tracking_mode_changed, DEFVAL(0));

    ClassDB::bind_method(D_METHOD("on_tracking_event_received", "frame"), &UltraleapDeviceNode::on_tracking_event_received);

    ClassDB::add_property(
        "UltraleapDeviceNode",
        PropertyInfo(
            Variant::NODE_PATH,
            "tracker",
            PROPERTY_HINT_NODE_TYPE,
            "UltraleapHandTracking"
        ),
        "set_tracker",
        "get_tracker"
    );

    ClassDB::add_property(
        "UltraleapDeviceNode",
        tracking_mode_info,
        "set_tracking_mode",
        "get_tracking_mode"
    );

    // Signals

    ClassDB::add_signal(
        "UltraleapDeviceNode",
        MethodInfo(
            "tracking_mode_changed",
            PropertyInfo(
                Variant::INT,
                "tracking_mode",
                PROPERTY_HINT_ENUM,
                "Desktop, HMD, Screentop"
            )
        )
    );

    ClassDB::add_signal(
        "UltraleapDeviceNode",
        MethodInfo(
            "frame_received",
            PropertyInfo(
                Variant::OBJECT,
                "frame",
                PROPERTY_HINT_RESOURCE_TYPE,
                "UltraleapFrame"
            )
        )
    );

    ClassDB::add_signal(
        "UltraleapDeviceNode",
        MethodInfo(
            "image_received",
            PropertyInfo(
                Variant::OBJECT,
                "frame",
                PROPERTY_HINT_RESOURCE_TYPE,
                "UltraleapImage"
            )
        )
    );
}

void UltraleapDeviceNode::set_tracker(NodePath value) {
    if (!tracker.is_empty()) {
        Node* tracking = get_node_or_null(tracker);
        if (tracking != NULL) {
            tracking->disconnect("device_added", Callable((Object*)this, "device_added"));
        }
    }

    Node* tracking = get_node_or_null(tracker);
    if (tracking != NULL) {
        tracking->connect("device_added", Callable((Object*)this, "device_added"));
    }
    tracker = value;
}

void UltraleapDeviceNode::_notification(int p_what) {
    if (p_what == Node::NOTIFICATION_READY) {
        UltraleapHandTracking* tracking = Object::cast_to<UltraleapHandTracking>(get_node_or_null(tracker));
        if (tracking != NULL) {
            tracking->connect("device_added", Callable((Object*)this, "device_added"));
            tracking->connect("device_removed", Callable((Object*)this, "device_removed"));
            for (int i = 0; i < tracking->devices->size(); i++) {
                Ref<UltraleapDevice> device = tracking->devices->get(i);
                if (!devices.has(device->serial)) {
                    devices.append(device->serial);
                }
            }
        }
    }
}

void UltraleapDeviceNode::_get_property_list(List<PropertyInfo> *p_list) const {
    p_list->push_back(
        PropertyInfo(
            Variant::INT,
            "device",
            PROPERTY_HINT_ENUM,
            String(",").join(devices).insert(0, "Any,")
        )
    );
}

bool UltraleapDeviceNode::_get(const StringName &p_name, Variant &r_ret) const {
    String name = p_name;
    if (name == "device") {
        r_ret = chosen_serial;
        return true;
    }
    return false;
}

bool UltraleapDeviceNode::_set(const StringName &p_name, const Variant &p_value) {
    String name = p_name;
    if (name == "device") {
        chosen_serial = p_value;
        notify_property_list_changed();
        return true;
    }
    return false;
}

UltraleapHandTracking* UltraleapDeviceNode::get_tracking() {
    return Object::cast_to<UltraleapHandTracking>(get_node_or_null(tracker));
}

void UltraleapDeviceNode::request_tracking_mode(UltraleapTypes::TrackingMode value) {
    if (current_device != NULL) {
        current_device->set_tracking_mode(value);
    }
}

void UltraleapDeviceNode::set_tracking_mode(UltraleapTypes::TrackingMode value) {
    tracking_mode = value;
}

void UltraleapDeviceNode::on_tracking_mode_changed(UltraleapTypes::TrackingMode value) {
    set_tracking_mode(value);
    emit_signal("tracking_mode_changed", value);
}

void UltraleapDeviceNode::on_tracking_event_received(Ref<UltraleapFrame> frame) {
    last_frame = frame;
    emit_signal("frame_received", last_frame);
}