#include "ultraleap.h"

#include <godot_cpp/core/class_db.hpp>

#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/classes/time.hpp>
#include <godot_cpp/variant/utility_functions.hpp>
#include <godot_cpp/variant/vector3.hpp>
#include <godot_cpp/classes/node3d.hpp>

#include <LeapC.h>
#include <time.h>

#include <thread>
#include <functional>

#include "frame.h"

using namespace godot;

UltraleapHandTracking::~UltraleapHandTracking() {
    if (_isRunning) {
        _isRunning = false;
        messageLoop.join();
    }
    LeapCloseConnection(connectionHandle);
    LeapDestroyConnection(connectionHandle);
}

void UltraleapHandTracking::_bind_methods() {
    PropertyInfo tracking_mode_info = PropertyInfo(
        Variant::INT,
        "tracking_mode",
        PROPERTY_HINT_ENUM,
        "Desktop, HMD, Screentop"
    );

    ClassDB::bind_method(D_METHOD("set_interpolate", "interpolate"), &UltraleapHandTracking::set_interpolate, DEFVAL(true));
    ClassDB::bind_method(D_METHOD("get_interpolate"), &UltraleapHandTracking::get_interpolate);
    ClassDB::add_property("UltraleapHandTracking", PropertyInfo(Variant::BOOL, "interpolate"), "set_interpolate", "get_interpolate");

    ClassDB::bind_method(D_METHOD("start"), &UltraleapHandTracking::start);

    ClassDB::bind_method(D_METHOD("set_last_frame", "last_frame"), &UltraleapHandTracking::set_last_frame);
    ClassDB::bind_method(D_METHOD("get_last_frame"), &UltraleapHandTracking::get_last_frame);

    ClassDB::bind_method(D_METHOD("set_tracking_mode", "tracking_mode"), &UltraleapHandTracking::set_tracking_mode, DEFVAL(0));
    ClassDB::bind_method(D_METHOD("get_tracking_mode"), &UltraleapHandTracking::get_tracking_mode);

    ClassDB::bind_method(D_METHOD("get_left_image"), &UltraleapHandTracking::get_left_image);
    ClassDB::bind_method(D_METHOD("get_right_image"), &UltraleapHandTracking::get_right_image);

    ClassDB::bind_method(D_METHOD("set_devices", "devices"), &UltraleapHandTracking::set_devices);
    ClassDB::bind_method(D_METHOD("get_devices"), &UltraleapHandTracking::get_devices);

    ClassDB::add_property(
        "UltraleapHandTracking",
        PropertyInfo(
            Variant::OBJECT,
            "last_frame",
            PROPERTY_HINT_RESOURCE_TYPE,
            "UltraleapFrame",
            PROPERTY_USAGE_READ_ONLY | PROPERTY_USAGE_EDITOR
        ),
        "set_last_frame",
        "get_last_frame"
    );

    ClassDB::add_property(
        "UltraleapHandTracking",
        tracking_mode_info,
        "set_tracking_mode",
        "get_tracking_mode"
    );

    ClassDB::add_property(
        "UltraleapHandTracking",
        PropertyInfo(
            Variant::ARRAY,
            "devices",
            PROPERTY_HINT_ARRAY_TYPE,
            "UltraleapDevice",
            PROPERTY_USAGE_READ_ONLY | PROPERTY_USAGE_EDITOR
        ),
        "set_devices",
        "get_devices"
    );

    // SIGNALS

    ClassDB::add_signal(
        "UltraleapHandTracking",
        MethodInfo(
            "tracking_mode_changed",
            tracking_mode_info
        )
    );

    ClassDB::add_signal(
        "UltraleapHandTracking",
        MethodInfo(
            "device_added",
            PropertyInfo(
                Variant::OBJECT,
                "device",
                PROPERTY_HINT_RESOURCE_TYPE,
                "UltraleapDevice"
            )
        )
    );

    ClassDB::add_signal(
        "UltraleapHandTracking",
        MethodInfo(
            "device_removed",
            PropertyInfo(
                Variant::OBJECT,
                "device",
                PROPERTY_HINT_RESOURCE_TYPE,
                "UltraleapDevice"
            )
        )
    );
}

void UltraleapHandTracking::start() {
    UtilityFunctions::print("Starting the tracking");
    LEAP_CONNECTION_CONFIG config;

    // Set connection to multi-device aware
    memset(&config, 0, sizeof(config));
    config.size = sizeof(config);
    config.flags = eLeapConnectionConfig_MultiDeviceAware;

    if(connectionHandle || LeapCreateConnection(&config, &connectionHandle) == eLeapRS_Success) {
        eLeapRS result = LeapOpenConnection(connectionHandle);
        if(result == eLeapRS_Success) {
            _isRunning = true;

            LeapSetPolicyFlags(connectionHandle, eLeapPolicyFlag_Images, 0);
            
            // Create the message poll thread
            messageLoop = std::thread(&UltraleapHandTracking::serviceMessageLoop, std::ref(*this));
        }
    }
}

bool UltraleapHandTracking::get_interpolate() {
    return interpolate;
}

void UltraleapHandTracking::set_interpolate(bool value) {
    interpolate = value;
}

void UltraleapHandTracking::handleConnectionEvent(const LEAP_CONNECTION_EVENT *evt) {
    IsConnected = true;
    UtilityFunctions::print("Connected");

    //LeapCreateClockRebaser(&clockSynchronizer);
}

void UltraleapHandTracking::serviceMessageLoop() {
    LEAP_CONNECTION_MESSAGE msg;
    eLeapRS result;
    int timeout = 100;

    while (_isRunning) {
        result = LeapPollConnection(connectionHandle, timeout, &msg);

        // Polling may have taken some time, re-check exit condition
        if (!_isRunning) {
            break;
        }

        switch (msg.type){
            case eLeapEventType_Connection:
                handleConnectionEvent(msg.connection_event);
                break;
            case eLeapEventType_ConnectionLost:
                break;
            case eLeapEventType_Device:
                handle_device_event(msg.device_event);
                break;
            case eLeapEventType_DeviceLost:
                handle_device_lost_event(msg.device_event);
                break;
            case eLeapEventType_DeviceFailure:
                UtilityFunctions::print("A device failed!");
                handle_device_lost_event(msg.device_event);
                break;
            case eLeapEventType_Tracking:
                // TODO: SECURE THIS WITH A MUTEX OR SOMETHING
                if(latest_frame == nullptr) {
                    latest_frame = new LEAP_TRACKING_EVENT;
                };

                // Store latest frame from any device as the latest frame
                // received by the system
                *latest_frame = *msg.tracking_event;

                // Iterate through the devices and find which device this frame
                // came from. Then store it in the device.
                for (int index = 0; index < devices->size(); index++) {
                    if (devices->get(index)->id == msg.device_id) {
                        devices->get(index)->on_frame_received(latest_frame);
                    }
                }
                
                if (devices->size() == 0) {
                    UltraleapFrame::fill_frame_data(last_frame, latest_frame);
                }
                else {
                    last_frame = devices->get(0)->last_frame;
                }
                break;
            case eLeapEventType_ImageComplete:
                break;
            case eLeapEventType_ImageRequestError:
                break;
            case eLeapEventType_LogEvent:
                break;
            case eLeapEventType_Policy:
                break;
            case eLeapEventType_ConfigChange:
                break;
            case eLeapEventType_ConfigResponse:
                break;
            case eLeapEventType_Image:
                // TODO: SECURE WITH A MUTEX
                *last_image_event = *msg.image_event;

                // Iterate through the devices and find which device this image
                // came from. Then store it in the device.
                for (int index = 0; index < devices->size(); index++) {
                    if (devices->get(index)->id == msg.device_id) {
                        //fill_images_data(devices->get(index)->left_image, devices->get(index)->right_image, msg.image_event->image);
                        devices->get(index)->on_image_received(last_image_event->image);
                    }
                }

                if (devices->size() == 0) {
                    UltraleapImage::fill_images_data(left_image, right_image, last_image_event->image);
                }
                else {
                    left_image = devices->get(0)->left_image;
                    right_image = devices->get(0)->right_image;
                }
                
                break;
            case eLeapEventType_PointMappingChange:
                break;
            case eLeapEventType_TrackingMode:
                // TODO: Check if having this and the policy event use the same path to notify a change in tracking mode
                handle_tracking_mode_event(msg.tracking_mode_event, msg.device_id);
                break;
            case eLeapEventType_LogEvents:
                break;
            case eLeapEventType_HeadPose:
                break;
            case eLeapEventType_IMU:
                break;
            case eLeapEventType_DeviceStatusChange:
                break;
            case eLeapEventType_DroppedFrame:
                break;
            case eLeapEventType_Eyes:
                break;
            case eLeapEventType_None:
                break;
            default:
                UtilityFunctions::print("Unhandled message type");
                break;
        }
    }
}

void UltraleapHandTracking::set_tracking_mode(UltraleapTypes::TrackingMode value) {
    if (value != tracking_mode) {
        uint64_t set_flag = 0;
        uint64_t clear_flag = 0;
        if (value == UltraleapTypes::TrackingMode::Desktop) {
            clear_flag = eLeapPolicyFlag_OptimizeHMD | eLeapPolicyFlag_OptimizeScreenTop;
        }
        else if (value == UltraleapTypes::TrackingMode::HMD) {
            set_flag = eLeapPolicyFlag_OptimizeHMD;
            clear_flag = eLeapPolicyFlag_OptimizeScreenTop;
        }
        else if (value == UltraleapTypes::TrackingMode::Screentop) {
            clear_flag = eLeapPolicyFlag_OptimizeHMD;
            set_flag = eLeapPolicyFlag_OptimizeScreenTop;
        }
        else {
            UtilityFunctions::print("Invalid value for tracking mode");
        }
        LeapSetPolicyFlags(connectionHandle, set_flag, clear_flag);
    }
}

void UltraleapHandTracking::handle_tracking_mode_event(const LEAP_TRACKING_MODE_EVENT* event, uint32_t device_id) {
    eLeapTrackingMode current_tracking_mode = event->current_tracking_mode;
    UltraleapTypes::TrackingMode tm = tracking_mode;

    // Parse into a defined type
    if (current_tracking_mode == eLeapTrackingMode_HMD) {
        tm = UltraleapTypes::TrackingMode::HMD;
    }
    else if (current_tracking_mode == eLeapTrackingMode_ScreenTop) {
        tm = UltraleapTypes::TrackingMode::Screentop;
    }
    else if (current_tracking_mode == eLeapTrackingMode_Desktop) {
        tm = UltraleapTypes::TrackingMode::Desktop;
    }

    if (device_id == 0) {
        if (tracking_mode != tm) {
            emit_signal("tracking_mode_changed", tm);
        }
    }
    else {
        Ref<UltraleapDevice> dev = devices->find_by_id(device_id);
        if (dev->tracking_mode != tm) {
            dev->tracking_mode_changed(tm);
        }
    }
}

void UltraleapHandTracking::handle_device_event(const LEAP_DEVICE_EVENT* event) {
    Ref<UltraleapDevice> dev = create_device(event->device);
    if (!devices->has_device(dev)) {
        devices->add_device(dev);
        emit_signal("device_added", dev);
    }
    else {
        UtilityFunctions::print("Device was already in the list, updating");
        devices->remove_device(dev);
        devices->add_device(dev);
        emit_signal("device_added", dev);
    }
}

void UltraleapHandTracking::handle_device_lost_event(const LEAP_DEVICE_EVENT* event) {
    UtilityFunctions::print("Device lost");

    Ref<UltraleapDevice> dev = create_device(event->device);
    if (devices->has_device(dev)) {
        devices->remove_device(dev);
        emit_signal("device_removed", dev);
    }
    else {
        UtilityFunctions::print("Device was not in the list");
    }
}

void UltraleapHandTracking::refresh_device_list() {
    UtilityFunctions::print("Refresh device list");

    uint32_t device_count = 0;
    eLeapRS result = LeapGetDeviceList(connectionHandle, NULL, &device_count);
    if (result != eLeapRS_Success) {
        UtilityFunctions::print("Error while counting the number of devices");
    }

    LEAP_DEVICE_REF* devices_ref = (LEAP_DEVICE_REF*)memalloc(device_count * sizeof(LEAP_DEVICE_REF));
    result = LeapGetDeviceList(connectionHandle, devices_ref, &device_count);

    if (result != eLeapRS_Success) {
        memfree(devices_ref);
        UtilityFunctions::print("Error while enumerating the devices");
    }

    devices->devices.clear();

    for (int index = 0; index < device_count; index++) {
        Ref<UltraleapDevice> dev = create_device(devices_ref[index]);
        devices->add_device(dev);
    }

    memfree(devices_ref);

    notify_property_list_changed();
}

Ref<UltraleapDevice> UltraleapHandTracking::create_device(LEAP_DEVICE_REF ref) {
    LEAP_DEVICE device;
    LeapOpenDevice(ref, &device);

    LEAP_DEVICE_INFO deviceInfo;
    memset(&deviceInfo, 0, sizeof(deviceInfo));

    // Use stack memory to allocate for the serial number field.
    char serial[1000];
    memset(serial, 0, sizeof(serial));
    deviceInfo.serial = serial;
    deviceInfo.serial_length = sizeof(serial) - 1;
    deviceInfo.size = sizeof(deviceInfo);

    LeapGetDeviceInfo(device, &deviceInfo);

    Ref<UltraleapDevice> dev = Ref<UltraleapDevice>(memnew(UltraleapDevice));
    dev->serial = String(deviceInfo.serial);
    dev->baseline = deviceInfo.baseline;
    dev->id = ref.id;

    dev->device_ref = ref;
    dev->connection = connectionHandle;

    UtilityFunctions::print("Created device with ID: ", ref.id, ", type: ", deviceInfo.pid, ", serial number: ", deviceInfo.serial);

    LeapCloseDevice(device);

    return dev;
}